/*
 * mailpg - a simple mail pager
 *
 * Copyright (C) 2010 Ali Gholami Rudi
 *
 * This file is released under GNU GPL version 2.
 */
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <termios.h>

#define BUFSIZE		(1 << 12)
#define MAXLINE		(1 << 7)
#define DEFROWS		25
#define DEFCOLS		80

#define COLOR_BLACK		0
#define COLOR_RED		1
#define COLOR_GREEN		2
#define COLOR_YELLOW		3
#define COLOR_BLUE		4
#define COLOR_MAGENTA		5
#define COLOR_CYAN		6
#define COLOR_WHITE		7
#define COLOR_DEFAULT		16

static char buf[BUFSIZE];
static int cur;
static int len;
static int tty;
static int rows, cols;
static int hdr = 1;		/* mail header lines */
static int cont;		/* last line is continued */

static int readk(void)
{
	char c = 0;
	if (read(tty, &c, 1) == -1)
		return -1;
	return c;
}

static void init_size(void)
{
	struct winsize win = {0};
	if (ioctl(tty, TIOCGWINSZ, &win) == -1) {
		rows = DEFROWS;
		cols = DEFCOLS;
	} else {
		rows = win.ws_row ? win.ws_row : DEFROWS;
		cols = win.ws_col ? win.ws_col : DEFCOLS;
	}
}

static int hdr_color(char *src, int len)
{
	switch (*src) {
	case 'S':
		if (src[1] == 'u')
			return COLOR_GREEN;
	case 'F':
		if (len > 4 && src[4] == ':')
			return COLOR_MAGENTA;
		break;
	case 'T':
		if (src[1] == 'o' && src[2] == ':')
			return COLOR_CYAN;
		break;
	}
	return COLOR_BLUE;
}

static int body_color(char *src, int len)
{
	switch (*src) {
	case '@':
		return COLOR_CYAN;
	case '-':
		return COLOR_RED;
	case '+':
		return COLOR_GREEN;
	case '>':
		return COLOR_GREEN;
	}
	return -1;
}

static void set_fg(int fg)
{
	char cmd[8];
	sprintf(cmd, "\33[%dm", fg >= 0 ? fg + 30 : 0);
	write(1, cmd, strlen(cmd));
}

static void draw_line(char *src, int len)
{
	static int fg;
	if (!cont)
		fg = hdr ? hdr_color(src, len) : body_color(src, len);
	if (*src == '\n')
		hdr = 0;
	if (fg != -1)
		set_fg(fg);
	write(1, src, len);
	if (fg != -1)
		set_fg(-1);
}

static int utf8len(int c)
{
	int l = 1;
	while (l < 6 && c & (0x40 >> l))
		l++;
	return l + 1;
}

static int esclen(char *s, int n)
{
	int l = 1;
	if (n > 1 && s[1] == '[') {
		l++;
		while (l < n && (s[l] < 64 || s[l] > 126))
			l++;
	}
	return l;
}

static int line_len(char *line, int llen)
{
	char *s = line;
	int w = 0;
	while (s - line < llen) {
		w++;
		if (*s == '\t')
			w = (w + 7) & ~7;
		if (*s == '\n') {
			s++;
			break;
		}
		if (w > cols)
			break;
		if (!(~*s & 0xc0)) {
			s += utf8len(*s);
			continue;
		}
		if (*s == '\33') {
			s += esclen(s, len - (s - buf));
			w--;
			continue;
		}
		s++;
	}
	return s - line;
}

static int put_line(void)
{
	int nr;
	int llen;
	if (len - cur < MAXLINE) {
		memmove(buf, buf + cur, len - cur);
		len = cur < len ? len - cur : 0;
		if ((nr = read(0, buf + len, sizeof(buf) - len)) >= 0)
			len += nr;
		if (!len)
			return 1;
		cur = 0;
	}
	llen = line_len(buf + cur, len - cur);
	draw_line(buf + cur, llen);
	cur += llen;
	cont = buf[cur - 1] != '\n';
	if (cont)
		write(1, "\n", 1);
	return 0;
}

static int put_lines(int n)
{
	int i;
	for (i = 0; i < n; i++)
		if (put_line())
			return -1;
	return 0;
}

static void put_cont(void)
{
	char *msg = "...";
	write(1, msg, strlen(msg));
}

int main(void)
{
	int quit = 0;
	tty = open("/dev/tty", O_RDONLY);
	init_size();
	if (put_lines(rows - 1))
		quit = 1;
	else
		put_cont();
	while (!quit) {
		switch (readk()) {
		case 'q':
			quit = 1;
			readk();
			break;
		case '\n':
			if (put_lines(rows - 1))
				quit = 1;
			else
				put_cont();
			break;
		}
	}
	close(tty);
	return 0;
}
