/*
 * see - a simple text pager
 *
 * Copyright (C) 2010 Ali Gholami Rudi
 *
 * This file is released under GNU GPL version 2.
 */
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include "term.h"

#define BUFSIZE		(1 << 20)
#define MAXLINE		(1 << 12)
#define DEFROWS		25
#define DEFCOLS		80
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define CTRLKEY(x)	((x) - 96)

static char buf[BUFSIZE];
static int len;
static int bol[MAXLINE];	/* offset of line beginnings */
static int nonl[MAXLINE];	/* nonzero for folded lines */
static int nbol;		/* the number of lines */

static int cur;
static int tty;
static int rows, cols;
static struct termios termios;
static char *filename = "-";

static int utf8len(int c)
{
	int l = 1;
	while (l < 6 && c & (0x40 >> l))
		l++;
	return l + 1;
}

static int esclen(char *s, int n)
{
	int l = 1;
	if (n > 1 && s[1] == '[') {
		l++;
		while (l < n && (s[l] < 64 || s[l] > 126))
			l++;
	}
	return l;
}

static char *putmem(char *d, char *s, int len)
{
	memcpy(d, s, len);
	return d + len;
}

static void die(char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

static void readall(int fd)
{
	int nr;
	while ((nr = read(fd, buf + len, sizeof(buf) - len)) > 0)
		len += nr;
	if (len == sizeof(buf))
		die("nomem: BUFSIZE reached!");
}

static void init_bols(void)
{
	int w = 0;
	char *s = buf;
	bol[nbol++] = 0;
	while (s < buf + len) {
		w++;
		if (nbol == MAXLINE)
			die("nomem: MAXLINE reached!");
		if (*s == '\t')
			w = (w + 7) & ~7;
		if (*s == '\n') {
			bol[nbol++] = s - buf + 1;
			w = 0;
		} else if (w > cols) {
			nonl[nbol - 1] = 1;
			bol[nbol++] = s - buf;
			w = 1;
		}
		if (!(~*s & 0xc0)) {
			s += utf8len(*s);
			continue;
		}
		if (*s == '\33') {
			s += esclen(s, len - (s - buf));
			w--;
			continue;
		}
		s++;
	}
	if (bol[nbol - 1] != len)
		bol[nbol] = len;
}

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) < (b) ? (b) : (a))
#define LINE(i)		(buf + bol[i])
#define LLEN(i)		(bol[(i) + 1] - bol[i])

static char *putline(char *s, int i)
{
	if (i >= 0 && i < nbol - 1) {
		s = putmem(s, LINE(i), LLEN(i));
		if (nonl[i])
			*s++ = '\n';
		return s;
	}
	return putmem(s, "~\n", 2);
}

static void printpage(void)
{
	char scr[1 << 12];
	char *s = scr;
	int i;
	*s++ = '\n';
	for (i = 0; i < rows; i++)
		s = putline(s, cur + i);
	write(1, scr, s - scr);
}

static void scroll(int n)
{
	char scr[1 << 12];
	char *s = scr;
	int i;
	if (cur + n - 1 < -rows)
		n = -rows - cur + 1;
	if (cur + n + 2 > nbol)
		n = nbol - cur - 2;
	if (n > 0) {
		cur += n;
		s = putmem(s, "\33[K", 3);
		for (i = MAX(cur, cur + rows - n); i < cur + rows; i++)
			s = putline(s, i);
	}
	if (n < 0) {
		n = -n;
		s = putmem(s, "\33[1H\33[K", 7);
		cur -= n;
		for (i = cur; i < MIN(cur + n, cur + rows); i++)
			s = putline(s, i);
		s += sprintf(s, "\33[%dH\33[K", rows + 1);
		printpage();
	}
	write(1, scr, s - scr);
}

static int readk(void)
{
	char c = 0;
	if (read(tty, &c, 1) == -1)
		return -1;
	return c;
}

static void init_size(void)
{
	struct winsize win = {0};
	if (ioctl(tty, TIOCGWINSZ, &win) == -1) {
		rows = DEFROWS;
		cols = DEFCOLS;
	} else {
		rows = win.ws_row ? win.ws_row : DEFROWS;
		cols = win.ws_col ? win.ws_col : DEFCOLS;
	}
	rows--;
}

static void term_setup(void)
{
	struct termios newtermios;
	tcgetattr(tty, &termios);
	newtermios = termios;
	newtermios.c_lflag &= ~ICANON;
	newtermios.c_lflag &= ~ECHO;
	tcsetattr(tty, TCSAFLUSH, &newtermios);
}

static void term_cleanup(void)
{
	tcsetattr(tty, 0, &termios);
}

static int arg;		/* commands' argument */

static int getarg(int def)
{
	int ret = arg ? arg : def;
	arg = 0;
	return ret;
}

static void loop(void)
{
	char msg[1 << 10];
	int quit = 0;
	int c;
	while (!quit) {
		switch ((c = readk())) {
		case 'q':
			quit = 1;
			break;
		case CTRLKEY('e'):
		case 'j':
		case '\n':
			scroll(getarg(1));
			break;
		case CTRLKEY('y'):
		case 'k':
			scroll(-getarg(1));
			break;
		case 'G':
			cur = getarg(nbol - rows) - 1;
			printpage();
			break;
		case '%':
			cur = nbol * getarg(0) / 100;
			printpage();
			break;
		case CTRLKEY('f'):
		case ' ':
			scroll(getarg(1) * rows);
			break;
		case CTRLKEY('l'):
			printpage();
			break;
		case CTRLKEY('b'):
		case 127:
			scroll(-getarg(1) * rows);
			break;
		case 27:
			arg = 0;
			break;
		case CTRLKEY('g'):
			sprintf(msg, "\r\33[K%s \t %d/%d \t %d%%\r",
				filename, cur + 1, nbol, cur * 100 / nbol);
			write(1, msg, strlen(msg));
			break;
		default:
			if (isdigit(c))
				arg = arg * 10 + c - '0';
		}
	}
}

int main(int argc, char *argv[])
{
	tty = open("/dev/tty", O_RDONLY);
	init_size();
	if (argc > 1) {
		int fd = open(argv[1], O_RDONLY);
		filename = argv[1];
		if (fd < 0)
			return 1;
		readall(fd);
		close(fd);
	} else {
		readall(0);
	}
	init_bols();
	term_setup();
	printpage();
	loop();
	term_cleanup();
	close(tty);
	return 0;
}
